-- Copyright (C) 2024 rstcxk
-- 
-- This program is free software: you can redistribute it and/or modify it under the terms of
-- the GNU Affero General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
-- without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more details.
-- 
-- You should have received a copy of the GNU Affero General Public License along with this program. If not, see <https://www.gnu.org/licenses/>. 

--- @classmod LanguageConstruct_UnquotedString
--	Describes a string that has no quotes enclosing it, supports escape sequences, custom terminators and escape sequences

local ParameterExpansion
local max_chars_in_table = 10000

local UnquotedString =
{
	instance_of = "LanguageConstruct_UnquotedString",
}

UnquotedString.__index = UnquotedString

function UnquotedString:new(t)
	if t == nil then
		t = {}
	end

	-- {{{ type checking
	if type(t) ~= "table" then error("t should be of type table. instead of " .. type(t)) end
	if t.text and type(t.text) ~= "string" then error("t.text should be of type string. instead of " .. type(t.text)) end
	-- }}}

	self = setmetatable(
	{
		elements = t.elements or {}
	}, UnquotedString)
	self.__index = self

	return self
end

function UnquotedString:evaluate(context)
	-- {{{ type checking
	if context.instance_of ~= "ShellContext" then error("context should be an instance of ShellContext. instead of " .. context.instance_of) end
	-- }}}

	local output_string = ""

	for _, v in pairs(self.elements) do
		if type(v) == "string" then
			output_string = output_string .. v
		elseif type(v) == "table" and v.instance_of == "LanguageConstruct_ParameterExpansion" then
			output_string = output_string .. v:evaluate(context)
		else
			error("unexpected type")
		end
	end

	return output_string
end

-- @tab terminator_set a dictionary containing letters that should terminate the string
function UnquotedString:parse(parser_context, terminator_set)
	-- {{{ type checking
	if parser_context.instance_of ~= "ParserContext" then error("parser_context should be an instance of ParserContext. instead of " .. parser_context.instance_of) end
	if type(terminator_set) ~= "table" then error("terminator_set should be of type table. instead of " .. type(terminator_set)) end
	-- }}}

	-- in lua the string concatentation operator is extreamly slow
	-- so instead of using it here, i append them to the char_table list
	-- and then table.concat() the list to get the final string
	--
	-- max_chars_in_table is the maximum of characters allowed in the table
	-- so that it wont use so much memory
	local char = parser_context:peek()
	local char_table = {}
	local char_table_index = 1
	local is_escaped = false
	local current_element = ""
	local parameter_expansion_instance

	while not (terminator_set[char] and not is_escaped) and not parser_context:is_EOF() do
		if not is_escaped and char == "$" then
			is_escaped = false

			ParameterExpansion = ParameterExpansion or require("language_constructs.parameter_expansion")

			current_element = current_element .. table.concat(char_table)
			char_table = {}
			char_table_index = 1
			table.insert(self.elements, current_element)
			current_element = ""


			parameter_expansion_instance = ParameterExpansion:new()
			parameter_expansion_instance:parse(parser_context)

			table.insert(self.elements, parameter_expansion_instance)
		elseif not is_escaped and char == "\\" then
			is_escaped = true
			parser_context:advance()
		else
			is_escaped = false

			char_table[char_table_index] = parser_context:consume()
			char_table_index = char_table_index + 1

			if char_table_index > max_chars_in_table then
				current_element = current_element .. table.concat(char_table)
				char_table = {}
				char_table_index = 1
			end
		end

		char = parser_context:peek()
	end

	current_element = current_element .. table.concat(char_table)
	if current_element ~= "" then
		-- handling edgecase where string ends with an expansion
		table.insert(self.elements, current_element)
	end
end

return UnquotedString
